{
    Component   : Range Calendar
    Description : This is a calendar component, where you can select days of a year
                  just like when you mark days on a real calendar.
    Author      : Ferenc Lendvai
    Version     : 1.0
    e-mail      : lendvai@honline.hu
    Legal issues: Copyright (C) 1998, 1999 by Ferenc Lendvai

                  This software is provided 'as-is', without any express or
                  implied warranty.  In no event will the author be held liable
                  for any  damages arising from the use of this software.

                  Permission is granted to anyone to use this software for any
                  purpose, including commercial applications, and to alter it
                  and redistribute it freely, subject to the following
                  restrictions:

               1. The origin of this software must not be misrepresented,
                  you must not claim that you wrote the original software.
                  If you use this software in a product, an acknowledgment
                  in the product documentation would be appreciated but is
                  not required.

               2. Altered source versions must be plainly marked as such, and
                  must not be misrepresented as being the original software.

               3. This notice may not be removed or altered from any source
                  distribution.
}

unit RngCalendar;

interface

uses
  Classes, Controls, Messages, Windows, Forms, Graphics, StdCtrls, Grids, SysUtils;

type
//  TRngDays is a non-visual day selection class. The calendar uses this underlaying class
//  to manager the information, and in this way it is possible to work with these data
//  without allocating any visual class.
    TRngDays     =  class
                    protected
                        m_Year : integer;
                        m_Days : array [1..366] of boolean;

                        procedure SetSelected(month, day: integer; v: boolean);
                        function  GetSelected(month, day: integer): boolean;

                        procedure SetItem(idx: integer; v: boolean);
                        function  GetItem(idx: integer): boolean;

                        procedure SetYear(year: integer);

                    public
                        constructor Create;

                        function  DaysPerMonth(AMonth: integer): integer;
                        function  IsLeapYear: boolean;

                        function  Days(AMonth, ADay: integer): integer;
                        function  IndexOf(dt: TDateTime): integer;

                        procedure SaveToStream(Stream: TStream);            // Used to store information (file, db)
                        procedure LoadFromStream(Stream: TStream);          // Just the opposite

                        procedure Clear;

                        procedure Assign(v: TObject);
                        function  IsEqual(v: TObject): boolean;

                        property  Items[i: integer]             : boolean read GetItem     write SetItem; default;
                        property  Selected[month, day: integer] : boolean read GetSelected write SetSelected;
                        property  Year                          : integer read m_Year      write SetYear;
                    end;

// The visual class
    TRngCalendar =  class(TCustomGrid)
                    protected
                        m_Group   : integer;
                        m_Days    : TRngDays;
                        m_IsDown  : boolean;
                        m_IsMoved : boolean;

                        procedure DrawCell(ACol, ARow: longint; ARect: TRect; AState: TGridDrawState); override;
                        function SelectCell(ACol, ARow: longint): boolean; override;
                        procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
                        procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
                        procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;

                        function  CellMonth(ACol, ARow: integer): integer;
                        function  CellDay(ACol, ARow: integer): integer;
                        function  CellDays(ACol, ARow: integer): integer;

                        procedure SetSelected(month, day: integer; v: boolean);
                        function  GetSelected(month, day: integer): boolean;

                        function  GetYear: integer;
                        procedure SetYear(year: integer);

                    public
                        constructor Create(AOwner: TComponent); override;
                        destructor  Destroy; override;

                        procedure SaveToStream(Stream: TStream);
                        procedure LoadFromStream(Stream: TStream);

                        property Selected[month, day: integer] : boolean read GetSelected write SetSelected;

                        property Days : TRngDays read m_Days;

                    published
                        property Year                          : integer read GetYear     write SetYear;
                        property GroupIndex                    : integer read m_Group     write m_Group;

                        property Align;
                        property BorderStyle;
                        property Color;
                        property Ctl3D;
                        property Col;
                        property Enabled;
                        property Font;
                        property ParentColor;
                        property ParentFont;
                        property ParentShowHint;
                        property PopupMenu;
                        property Row;
                        property ShowHint;
                        property TabOrder;
                        property TabStop;
                        property Visible;

                        property OnClick;
                        property OnDblClick;
                        property OnDragDrop;
                        property OnDragOver;
                        property OnEndDrag;
                        property OnEnter;
                        property OnExit;
                        property OnKeyDown;
                        property OnKeyPress;
                        property OnKeyUp;
                        property OnStartDrag;
                    end;

{$IFNDEF USE_COMMON_REG}
procedure Register;
{$ENDIF}

implementation

constructor TRngDays.Create;
var
    y, m, d: word;
begin
    DecodeDate(Date, y, m, d);
    SetYear(y)
end;

procedure TRngDays.Clear;
var
    i : integer;
begin
    for i := 1 to 366 do
        m_Days[i] := false
end;

procedure TRngDays.Assign(v: TObject);
var
    i : integer;
begin
    if v is TRngDays then
        for i := 1 to 366 do
            m_Days[i] := TRngDays(v).m_Days[i]
end;

function TRngDays.IsEqual(v: TObject): boolean;
var
    i : integer;
begin
    Result := false;
    if v is TRngDays then
        for i := 1 to 366 do
            if m_Days[i] <> TRngDays(v).m_Days[i] then
                Exit;
    Result := true
end;

function TRngDays.IsLeapYear: boolean;
begin
    Result := (m_Year mod 4 = 0) and ((m_Year mod 100 <> 0) or (m_Year mod 400 = 0))
end;

function TRngDays.DaysPerMonth(AMonth: integer): integer;
const
    DaysInMonth: array[1..12] of Integer = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
begin
    Result:= DaysInMonth[AMonth];
    if (AMonth = 2) and IsLeapYear then
        Inc(Result)
end;

procedure TRngDays.SetYear(year: integer);
begin
    m_Year := year;
    Clear
end;

function TRngDays.Days(AMonth, ADay: integer): integer;
var
    i : integer;
begin
    Result := ADay;
    for i := 1 to AMonth - 1 do
        Result := Result + DaysPerMonth(i)
end;

function TRngDays.IndexOf(dt: TDateTime): integer;
var
    y, m, d : word;
begin
    DecodeDate(dt, y, m, d);
    Assert(y = m_Year, 'Different year: ' + IntToStr(y) + ' <> ' + IntToStr(m_Year));
    Result := Days(m, d)
end;

procedure TRngDays.SaveToStream(Stream: TStream);
begin
    Stream.Write(m_Days, SizeOf(m_Days))
end;

procedure TRngDays.LoadFromStream(Stream: TStream);
begin
    Stream.Read(m_Days, SizeOf(m_Days))
end;

procedure TRngDays.SetItem(idx: integer; v: boolean);
begin
    Assert((1 <= idx) and (idx <= 366), 'Bad day index: ' + IntToStr(idx));
    m_Days[idx] := v
end;

function TRngDays.GetItem(idx: integer): boolean;
begin
    Assert((1 <= idx) and (idx <= 366), 'Bad day index: ' + IntToStr(idx));
    Result := m_Days[idx]
end;

procedure TRngDays.SetSelected(month, day: integer; v: boolean);
begin
    m_Days[Days(month, day)] := v
end;

function TRngDays.GetSelected(month, day: integer): boolean;
begin
    Result := m_Days[Days(month, day)]
end;

constructor TRngCalendar.Create(AOwner: TComponent);
begin
    inherited;

    m_Days           := TRngDays.Create;

    m_IsDown         := false;
    m_IsMoved        := false;
    ColCount         := 7 * 12;
    RowCount         := 8;
    FixedCols        := 1;
    FixedRows        := 1;
    FixedColor       := clWindow;
    DefaultColWidth  := 18;
    DefaultRowHeight := 18;
    Font.Size        := 8;
    RowHeights[0]    := 22;
    ColWidths[0]     := 29;
    Options          := [];
    Width            := 400;
    Height           := 169;
end;

destructor TRngCalendar.Destroy;
begin
    if m_Days <> nil then
        m_Days.Free;
    inherited
end;

function TRngCalendar.GetYear: integer;
begin
    Result := m_Days.Year
end;

procedure TRngCalendar.SetYear(year: integer);
var
    i : integer;
begin
    m_Days.Year := year;
    Col := 1;
    Row := DayOfWeek(EncodeDate(Year, 1, 1)) - 1;
    Invalidate
end;

function TRngCalendar.CellMonth(ACol, ARow: integer): integer;
begin
    Result := (ACol - 1) div 7 + 1
end;

function TRngCalendar.CellDay(ACol, ARow: integer): integer;
var
    ho     : integer;
    offset : integer;
begin
    ho := CellMonth(ACol, ARow);
    offset := DayOfWeek(EncodeDate(Year, ho, 1)) - 1;
    if offset = 0 then
        offset := 7;
    Result  := (ACol - (ho - 1) * 7 - 1) * 7 + ARow - offset + 1;
    if (Result < 1) or (Result > m_Days.DaysPerMonth(ho)) then
        Result := 0
end;

function TRngCalendar.CellDays(ACol, ARow: integer): integer;
var
    ho  : integer;
    nap : integer;
begin
    ho  := CellMonth(ACol, ARow);
    nap := CellDay(ACol, ARow);
    Result := m_Days.Days(ho, nap)
end;

procedure TRngCalendar.DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);
const
  DaysInMonth: array[1..12] of Integer = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
var
    s      : string;
    s1     : string;
    i      : integer;
    ho     : integer;
    nap    : integer;
    r      : TRect;
begin
    if ACol = 0 then
    begin
        if (1 <= ARow) and (ARow <= 7) then
        begin
            i := ARow;
            if i = 7 then
                i := 1
            else
                i := i + 1;
            s := ShortDayNames[i];
            s1 := s[1];
            s1 := AnsiUpperCase(s1);
            s[1] := s1[1];
            if ARow < 6 then
                Canvas.Font.Color := clNavy
            else
                Canvas.Font.Color := clRed;
            Canvas.Font.Style := Canvas.Font.Style + [fsBold];
            DrawText(Canvas.Handle, PChar(s), Length(s), ARect, DT_LEFT or DT_VCENTER or DT_SINGLELINE)
        end
    end
    else if ARow = 0 then
    begin
        s := LongMonthNames[(ACol -1) div 7 + 1];
        s1 := s[1];
        s1 := AnsiUpperCase(s1);
        s[1] := s1[1];
        Canvas.Font.Color := clNavy;
        Canvas.Font.Style := Canvas.Font.Style + [fsBold];
        i := ((ACol -1) div 7) * 7 + 1;
        r := ARect;
        r.Left  := r.Left - (ACol - i) * DefaultColWidth;
        r.Right := r.Left + 6 * DefaultColWidth;
        DrawText(Canvas.Handle, PChar(s), Length(s), r, DT_CENTER or DT_VCENTER or DT_SINGLELINE);
        Canvas.Pen.Color := clBlack;
        Canvas.MoveTo(r.Left,  r.Bottom - 1);
        Canvas.LineTo(r.Right, r.Bottom - 1)
    end
    else if (ACol - 1) mod 7 <> 6 then
    begin
        ho   := CellMonth(ACol, ARow);
        nap  := CellDay(ACol, ARow);
        if nap <> 0 then
        begin
            s := IntToStr(nap);
            if ARow < 6 then
                Canvas.Font.Color := clBlack
            else
                Canvas.Font.Color := clRed;
            DrawText(Canvas.Handle, PChar(s), Length(s), ARect, DT_CENTER or DT_VCENTER or DT_SINGLELINE);
            if m_Days.Selected[ho, nap] then
            begin
                Canvas.Pen.Color := clRed;
                Canvas.MoveTo(ARect.Left,  ARect.Top);
                Canvas.LineTo(ARect.Right, ARect.Bottom);
                Canvas.MoveTo(ARect.Left,  ARect.Bottom);
                Canvas.LineTo(ARect.Right, ARect.Top)
            end
        end
    end
end;

function TRngCalendar.SelectCell(ACol, ARow: longint): boolean;
var
    ho, nap : integer;
begin
    Result := inherited SelectCell(ACol, ARow) and (CellDay(ACol, ARow) <> 0);
    if Result {and m_IsMoved} then
    begin
        ho   := CellMonth(ACol, ARow);
        nap  := CellDay(ACol, ARow);
        m_Days.Selected[ho, nap] := not m_Days.Selected[ho, nap]
    end
end;

procedure TRngCalendar.SaveToStream(Stream: TStream);
begin
    m_Days.SaveToStream(Stream)
end;

procedure TRngCalendar.LoadFromStream(Stream: TStream);
begin
    m_Days.LoadFromStream(Stream);
    Invalidate
end;

procedure TRngCalendar.SetSelected(month, day: integer; v: boolean);
begin
    m_Days.Selected[month, day] := v
end;

function TRngCalendar.GetSelected(month, day: integer): boolean;
begin
    Result := m_Days.Selected[month, day]
end;

procedure TRngCalendar.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
    inherited;

    m_IsDown  := true;
    m_IsMoved := false
end;

procedure TRngCalendar.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
    inherited;

    if m_IsDown then
        m_IsMoved := true
end;

procedure TRngCalendar.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
    inherited;

    m_IsDown  := false;
    m_IsMoved := false
end;

{$IFNDEF USE_COMMON_REG}
procedure Register;
begin
    RegisterComponents('LF Pages', [TRngCalendar])
end;
{$ENDIF}

end.

